/*
Copyright (C) 2014 Graham Jones, www.indriid.com
*/

package stacey;

import beast.core.*;
import beast.core.parameter.RealParameter;
import beast.evolution.alignment.TaxonSet;
import beast.evolution.tree.TreeDistribution;
import stacey.debugtune.Checks;
import stacey.util.Bindings;
import stacey.util.InverseGammaMixture;
import stacey.util.PIOProbabilityGeneTree;
import stacey.util.PIOProbabilitySMCTree;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

/**
 * Created by Graham Jones on 19/08/2014.
 */

/*
This is the central class of STACEY (if any is).

It calculates the coalescent probability.
It is complicated because the calculation is quite complicated
and because it does some optimization to avoid recalculating
everything when a single gene tree is changed, and to avoid calculating
a likelihood when one or more gene trees are incompatible.
*/

/*
General notes on code organization in STACEY. 2014-11-04

Package stacey contains classes derived from BEASTObject. They all
have Inputs. These are top level classes constructed from XML.

Package stacey.util contains everything else. None of these classes
calls methods in stacey.

                                 In stacey
extend Distribution:
    BirthDeathCollapseModel, PIOMSCoalescentDistribution
extend CalculationNode:
   GtreeAndCoalFactor
extend Operator:
    CoordinatedPruneRegraft, FocusedNodeHeightScaler, NodesNudge
extend BEASTObject implementing Loggable:
    BirthDeathCollapseNClustersStatistic, PopSampleStatistic
extend BEASTObject:
    InverseGammaComponent
extend InputEditor.Base
    GtreeAndCoalFactorInputEditor, PIOMSCoalescentDistributionInputEditor

The operators and this class, PIOMSCoalescentDistribution, are quite big, the rest small.

The editors aren't working.


                         In stacey.util - small classes

Bindings is a small class which stores the information in a TaxonSet of TaxonSets of Taxons,
and provides various access methods. It is an immutable singleton though I haven't enforced this.

BitUnion is a low-level utility class providing a minimal 'Bitset' implementation.

FSMoveSMCNodeInfo, FSMoveSMCNodeUnions, NudgeMoveSMCNodeInfo, InverseGammaMixture
are small immutable value classes for passing information around.

Misc is formatting for debug output

                         In stacey.util - bigger classes

OperatorsGeneTree, OperatorsSMCTree, PIOProbabilityGeneTree, PIOProbabilitySMCTree
do a lot of the work in STACEY. They each represent auxiliary information
associated with a BEAST tree, especially unions of minimal clusters. They have a lot
in common and perhaps should be derived from something.

They are constructed initially using constructors which claim memory for empty
structures. They have update() methods to be called before doing calculations with them.

The pair for operators are constructed by each operator separately. They are updated
before doing the move, and have methods reset() which are called after the move,
which takes them back to the empty structure. Eg:
    public double proposal() {
        update(); // calls update() on OperatorsSMCTree and all OperatorsGeneTree's.
        double logHR = doFocusedScalerMove();
        reset();  // calls reset() on OperatorsSMCTree and all OperatorsGeneTree's.
        return logHR;
    }
They do not maintain information over MCMC generations.

The pair for probability calculations are constructed by PIOMSCoalescentDistribution
(this class) initially.

PIOProbabilitySMCTree maintains information over MCMC generations;
it is updated when out of date. It is a bigger structure since it
maintains information about coalescences in all gene trees at each branch. This
therefore needs to be stored and restored. It has a replaceWith() method to
support this.

PIOProbabilityGeneTree is smaller but there is (of course) one for each gene tree.
These are updated on a per-calculation basis. As things are on 2014-11-04,
every call to fitsInSpeciesTree(P) and fillSpeciesTreeWithGeneCoalescentInfo()
calls a private update() before doing its stuff. The contents of a
PIOProbabilityGeneTree don't need to be stored and restored.



*/





@Description("The smcCoalescent distribution for the probability of all gene trees conditioned on a species tree or minimal clusters tree. <br/>")

@Citation(value = "Graham Jones (www.indriid.com). " +
        "\'STACEY: species delimitation and phylogeny estimation under the multispecies coalescent\'. " +
        "http://biorxiv.org/content/early/2014/10/09/010199",
        firstAuthorSurname = "Jones")


public class PIOMSCoalescentDistribution extends TreeDistribution {

    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<List<GtreeAndCoalFactor>> geneTreesInput =
            new Input<List<GtreeAndCoalFactor>>("geneTree",
                    "All gene trees",
                    new ArrayList<GtreeAndCoalFactor>());

    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<List<InverseGammaComponent>> priorComponentsInput =
            new Input<List<InverseGammaComponent>>("popPriorInvGamma",
                    "Component of mixture of inverse gamma distributions used as a prior for the per-branch populations",
                    new ArrayList<InverseGammaComponent>(),
                    Input.Validate.REQUIRED);

    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<RealParameter> popPriorScale =
            new Input<RealParameter>("popPriorScale",
                    "Overall scale for population size",
                    Input.Validate.REQUIRED);

    // Needed for Beauti template, but not used here
    @SuppressWarnings("UnusedDeclaration")
    public Input<TaxonSet> taxonSetInput =
            new Input<TaxonSet>("taxonset",
                    "set of taxa mapping lineages to species", Input.Validate.REQUIRED);


    // these are initialised once and not changed, but can't be final cos of
    // initAndValidate() business. (Internally pioGeneTrees changes a lot as gene trees change.)
    private PIOProbabilityGeneTree[] pioGeneTrees;


    // This is the data that is stored and restored.
    // smcTree stores intermediate probability calculations per-gene at each branch.
    // gTreeFits records which gene trees are compatible.
    // The per-gene info in smcTree is only updated when all gene trees are compatible,
    // and the compatibility is not recalculated beyond finding one incompatible gene tree.
    // The two arrays of dirty flags are used to deal with this.
    private PIOProbabilitySMCTree smcTree;
    private boolean [] gTreeFits;
    private boolean smcTreeIsDirty;
    private boolean [] gTreeFitIsDirty;
    private boolean [] gTreeLhoodInfoIsDirty;

    private PIOProbabilitySMCTree oldsmcTree;
    private boolean [] oldgTreeFits;
    private boolean oldsmcTreeIsDirty;
    private boolean [] oldgTreeFitIsDirty;
    private boolean [] oldgTreeLhoodInfoIsDirty;


    private boolean debugFlag = Boolean.valueOf(System.getProperty("stacey.debug"));
    private int numberofdebugchecks = 0;
    private final static int maxnumberofdebugchecks = 1000000;



    InverseGammaMixture getInverseGammaMixture() {
        int n = priorComponentsInput.get().size();
        InverseGammaMixture igm = new InverseGammaMixture(n);
        for (int c = 0; c < n; c++) {
            igm.setWeight(c, priorComponentsInput.get().get(c).getWeight());
            igm.setAlpha(c, priorComponentsInput.get().get(c).getAlpha());
            igm.setBeta(c, priorComponentsInput.get().get(c).getBeta());
        }
        return igm;
    }



    @Override
    public void initAndValidate() throws Exception {
        super.initAndValidate(); // ??

        List<GtreeAndCoalFactor> gTreeCFs = geneTreesInput.get();
        final int nGeneTrees = gTreeCFs.size();
        int [] geneTreeTipCounts = new int[nGeneTrees];
        for (int j = 0; j < geneTreeTipCounts.length; j++) {
            geneTreeTipCounts[j] = gTreeCFs.get(j).gtree.get().getLeafNodeCount();
        }
        Bindings bindings = new Bindings(treeInput.get().getTaxonset());
        smcTree = new PIOProbabilitySMCTree(treeInput.get(), bindings, geneTreeTipCounts);
        oldsmcTree = new PIOProbabilitySMCTree(treeInput.get(), bindings, geneTreeTipCounts);
        pioGeneTrees = new PIOProbabilityGeneTree[nGeneTrees];
        // make the PIOProbabilityGeneTrees
        for (int j = 0; j < geneTreesInput.get().size(); j++) {
            pioGeneTrees[j] = new PIOProbabilityGeneTree(gTreeCFs.get(j).getTree(),
                    gTreeCFs.get(j).getCoalFactor(), bindings, j);
        }
        // normalize the inverse gamma component weights to sum to 1
        double totalweight = 0.0;
        for (int i = 0; i < priorComponentsInput.get().size(); i++) {
            totalweight += priorComponentsInput.get().get(i).getWeight();
        }
        for (int i = 0; i < priorComponentsInput.get().size(); i++) {
            priorComponentsInput.get().get(i).normalizeWeight(totalweight);
        }
        // set up all the dirtiness flags
        gTreeFits =    new boolean[nGeneTrees];
        oldgTreeFits = new boolean[nGeneTrees];

        smcTreeIsDirty =    true;
        oldsmcTreeIsDirty = true;

        gTreeFitIsDirty =    new boolean[nGeneTrees];
        oldgTreeFitIsDirty = new boolean[nGeneTrees];

        gTreeLhoodInfoIsDirty =    new boolean[nGeneTrees];
        oldgTreeLhoodInfoIsDirty = new boolean[nGeneTrees];
        for (int j = 0; j < pioGeneTrees.length; j++) {
            gTreeFitIsDirty[j] =    true;
            oldgTreeFitIsDirty[j] = true;
            gTreeLhoodInfoIsDirty[j]    = true;
            oldgTreeLhoodInfoIsDirty[j] = true;
        }
    }



    @Override
    public double calculateLogP() {
        // Basically, this calls logLhoodAllGeneTreesInSMCTree(); the rest is debug checks


        numberofdebugchecks++;
        if (numberofdebugchecks > maxnumberofdebugchecks) {
            debugFlag = false;
        }
        double fastLogP = logLhoodAllGeneTreesInSMCTree(getInverseGammaMixture(),
                popPriorScale.get().getValue(), false);

        if (debugFlag) {
            double robustLogP = logLhoodAllGeneTreesInSMCTree(getInverseGammaMixture(),
                    popPriorScale.get().getValue(), true);
            if (Math.abs(fastLogP - robustLogP) > 1e-12) {
                System.out.println("BUG in calculateLogP() in PIOMSCoalescentDistribution");
                assert false;
            }
        }
        // should I be changing the member logP of Distribution?
        // Yes, say Remco and alexei, until a calculate-and-cache is implemented.
        logP = fastLogP;
        return logP;
    }



    @Override
    protected boolean requiresRecalculation() {
        smcTreeIsDirty = treeInput.get().somethingIsDirty();
        for (int j = 0; j < pioGeneTrees.length; j++) {
            boolean gTreeDirty = pioGeneTrees[j].tree.somethingIsDirty();
            // I don't update everything in case of incompatibility, so need to OR
            gTreeFitIsDirty[j] |= gTreeDirty;
            gTreeLhoodInfoIsDirty[j] |=  gTreeDirty;
        }
        return true;

        /* I think its rare that neither smcTree nor any gene tree changed,
        so might as well return true always.
        The code below might be more efficient. What about origin height?
        boolean somethingDirty = false;
        smcTreeIsDirty = treeInput.get().somethingIsDirty();
        somethingDirty |= smcTreeIsDirty;
        for (int j = 0; j < pioGeneTrees.length; j++) {
            boolean gTreeDirty = pioGeneTrees[j].tree.somethingIsDirty();
            // I don't update everything in case of incompatibility, so need to OR
            gTreeFitIsDirty[j] |= gTreeDirty;
            gTreeLhoodInfoIsDirty[j] |=  gTreeDirty;
            somethingDirty |= gTreeFitIsDirty[j];
            somethingDirty |= gTreeLhoodInfoIsDirty[j];
        }
        somethingDirty |= popPriorScale.get().somethingIsDirty();
        return somethingDirty;*/
    }


    @Override
    public List<String> getArguments() {
        return null;
    }


    @Override
    public List<String> getConditions() {
        return null;
    }


    @Override
    public void sample(final State state, final Random random) {
    }


    @Override
    public void store() {
        oldsmcTree.replaceWith(smcTree);
        System.arraycopy(gTreeFits, 0, oldgTreeFits, 0, oldgTreeFits.length);
        oldsmcTreeIsDirty = smcTreeIsDirty;
        System.arraycopy(gTreeFitIsDirty, 0, oldgTreeFitIsDirty, 0, oldgTreeFitIsDirty.length);
        System.arraycopy(gTreeLhoodInfoIsDirty, 0, oldgTreeLhoodInfoIsDirty, 0, oldgTreeLhoodInfoIsDirty.length);
        super.store();
    }



    @Override
    public void restore() {
        PIOProbabilitySMCTree tmp = smcTree;
        smcTree = oldsmcTree;
        oldsmcTree = tmp;
        System.arraycopy(oldgTreeFits, 0, gTreeFits, 0, gTreeFits.length);
        smcTreeIsDirty = oldsmcTreeIsDirty;
        System.arraycopy(oldgTreeFitIsDirty, 0, gTreeFitIsDirty, 0, gTreeFitIsDirty.length);
        System.arraycopy(oldgTreeLhoodInfoIsDirty, 0, gTreeLhoodInfoIsDirty, 0, gTreeLhoodInfoIsDirty.length);
        super.restore();
    }




    private double logLhoodAllGeneTreesInSMCTree(InverseGammaMixture igm,
                                                 double popPriorScale, boolean robust) {
        if (robust) {
            smcTreeIsDirty = true;
        }
        // if smcTree has changed everything is wrong, so make everything
        // dirty then update the smcTree
        if (smcTreeIsDirty) {
            for (int j = 0; j < pioGeneTrees.length; j++) {
                gTreeFitIsDirty[j] = gTreeLhoodInfoIsDirty[j] = true;
            }
            smcTree.update();
            smcTreeIsDirty = false;
        }
        // unless gtree fit is clean and the gtree fits, lhood dirtyness may be wrong
        for (int j = 0; j < pioGeneTrees.length; j++) {
            if (gTreeFitIsDirty[j]  ||  !gTreeFits[j]) {
                gTreeLhoodInfoIsDirty[j] = true;
            }
        }
        // make gTreeFits up to date
        for (int j = 0; j < pioGeneTrees.length; j++) {
            if (gTreeFitIsDirty[j]) {
                gTreeFits[j] = pioGeneTrees[j].fitsInSpeciesTree(smcTree);
                gTreeFitIsDirty[j] = false;
            } else {
                if (debugFlag) {
                    if (gTreeFits[j] != pioGeneTrees[j].fitsInSpeciesTree(smcTree)) {
                        System.out.println("BUG in logLhoodAllGeneTreesInSMCTree() gTreeFits[j] wrong");
                        assert false;
                    }
                }
            }
        }
        // if incompatibility, return -oo
        boolean allGtreesFit = true;
        for (int j = 0; j < pioGeneTrees.length; j++) {
            allGtreesFit = allGtreesFit && gTreeFits[j];
        }
        if (!allGtreesFit) {
            return Double.NEGATIVE_INFINITY;
        }
        // find max height of any gene tree for lhood calc
        double maxgtreehgt = 0.0;
        for (GtreeAndCoalFactor gtcf : geneTreesInput.get()) {
            maxgtreehgt = Math.max(maxgtreehgt, gtcf.getTree().getRoot().getHeight());
        }
        // update the per-gene info in smcTree branches where needed
        for (int j = 0; j < pioGeneTrees.length; j++) {
            if (gTreeLhoodInfoIsDirty[j]) {
                pioGeneTrees[j].fillSpeciesTreeWithGeneCoalescentInfo(smcTree, maxgtreehgt);
                gTreeLhoodInfoIsDirty[j] = false;
            } else {
                if (debugFlag) {
                    Checks.debugCheckCoalescences(smcTree, pioGeneTrees[j], j);
                }
            }
        }
        // calculate and sanity-check the result
        double logPGS = smcTree.logProbAllGeneTreesInSMCTree(igm, popPriorScale);
        assert !Double.isNaN(logPGS);
        assert !Double.isInfinite(logPGS);
        return logPGS;
    }





}
